local Recount2 = Recount2
local SM = LibStub:GetLibrary("LibSharedMedia-3.0")

local Table={}

local tremove = table.remove
local tinsert = table.insert
local table_sort = table.sort
local math_floor = math.floor
local math_max = math.max
local math_min = math.min

local function SetFontSize(string, size)
	local Font, Height, Flags = string:GetFont()
	string:SetFont(Font, size, Flags)
end

local function RowSizeChanged(self)
	local width = self:GetWidth()
	if width<self.nameSpace then
		self.NameText:SetPoint("RIGHT", self, "LEFT", width, 0)
	else
		self.NameText:SetPoint("RIGHT", self, "LEFT", self.nameSpace, 0)
	end
	self:UpdateColumns()
end

local function SetNameSpace(self, width)
	self.nameSpace = width
	self:OnSizeChanged()
end

local function RowSetOffset(self, offset)
	self.offset = offset
	self:UpdateColumns()
end

local function AddColumn(self, width, pos)
	local fontString
	if #self.extraStrings>0 then
		fontString = tremove(self.extraStrings)
	else
		fontString = self.StatusBar:CreateFontString(nil ,"OVERLAY","GameFontHighlightSmall")
		Recount2:AddFontString(fontString)
		SetFontSize(fontString, self:GetHeight()*0.8)
	end
	
	if pos==nil then
		pos = #self.columns + 1
	end

	fontString:SetJustifyH("CENTER")
	tinsert(self.columns, pos, fontString)
	tinsert(self.columnWidth, pos, width)

	self:UpdateColumns()
end

local function CenterLabels(self)
	for _, v in ipairs(self.columns) do
		v:SetJustifyH("CENTER")
	end
end

local function SetName(self, text)
	self.NameText:SetText(text)
end

local function SetColumn(self, num, text)
	self.columns[num]:SetText(text)
end

local function SetColumnWidth(self, num, width)
	self.columnWidth[num] = width
	self:UpdateColumns()
end

local function UpdateColumns(self)
	local offset = self.offset - self.nameSpace 
	local leftBounds = self.nameSpace
	local rightBounds 
	if self:GetRight() then
		rightBounds = self:GetRight()-self:GetLeft()
	else
		rightBounds = self:GetWidth()
	end
	local left, right, width
	local columnWidth = self.columnWidth
	for k, column in ipairs(self.columns) do
		column:ClearAllPoints()
		width = columnWidth[k]
		left = -offset
		right = width-offset
		if right<leftBounds or left>rightBounds then
			column:Hide()
		else
			if left<leftBounds then
				left = leftBounds
			end
			if right>rightBounds then
				right = rightBounds
			end
			column:SetPoint("LEFT", self, "LEFT", left, 0)
			column:SetPoint("RIGHT", self, "LEFT", right, 0)

			column:Show()
		end
		offset = offset-width-4
	end
end

local function Row_MouseOnColumn(self)
	local x, y = GetCursorPosition()
	local scale = self:GetEffectiveScale() 
	x = x/scale-self:GetLeft()
	y = y/scale-self:GetBottom()

	local w = self:GetRight() - self:GetLeft()
	local h = self:GetTop() - self:GetBottom()

	if y<0 or y>h or x<0 or x>w then
		return nil
	end

	local nameSpace = self.nameSpace

	if x < nameSpace then
		return 0
	end

	x = x - nameSpace + self.offset

	for i, colWidth in ipairs(self.columnWidth) do
		if x < colWidth then
			return i
		end

		x = x - colWidth
	end

	return -1
end

local function Row_MouseDown(self)
	if self.callback then
		self.callback(self:GetParent(), "OnMouseDown", arg1, self.rowID, self:MouseOnColumn())
	end
end

local function Row_MouseUp(self)
	if self.callback then
		self.callback(self:GetParent(), "OnMouseUp", arg1, self.rowID, self:MouseOnColumn())
	end
end

local function TooltipPopup(self)
	self.tooltipTimerGoing = nil
	self.tooltipShowing = true

	self.tooltipFunc(unpack(self.tooltipArgs))
end

local function Row_MouseEnter(self)
	--need to start a timer to popup a tooltip
	if self.tooltipFunc then
		Recount2:ScheduleTimer("Tooltip", TooltipPopup, 1, self)
		self.tooltipTimerGoing = true
	end
	
	if self.highlight then
		self.Highlight:Show()
	end
end

local function Row_MouseLeave(self)
	if self.tooltipTimerGoing then
		Recount2:RemoveTimer("Tooltip")
		self.tooltipTimerGoing = false
	end

	if self.highlight then
		self.Highlight:Hide()
	end

	if self.tooltipShowing then
		Recount2.Widgets:DropdownClear()
		self.tooltipShowing = nil
	end
end

local function Row_SetTooltipFunc(self, func, args)
	self.tooltipFunc = func
	self.tooltipArgs = args

	--If showing need to update!
	if self.tooltipShowing then
		func(unpack(args))
	end
end

local function RemoveColumn(self, num)
	if num<1 or num>#self.columns then
		return
	end
	
	self.columns[num]:Hide()
	tinsert(self.extraStrings, self.columns[num])
	tremove(self.columns, num)

	self:UpdateColumns()
end

local function RemoveColumns(self)
	for k, v in ipairs(self.columns) do
		v:Hide()
		tinsert(self.extraStrings, v)
		self.columns[k] = nil		
	end
end

local function CreateRow(parent, rowWidth, rowHeight, nameSpace, rowID, callback)
	local row=CreateFrame("Button", nil, parent)
	row:SetWidth(rowWidth)
	row:SetHeight(rowHeight)
	row.rowID = rowID
	row.callback = callback

	--Create Bar
	row.StatusBar = CreateFrame("StatusBar",nil,row)
	row.StatusBar:SetAllPoints(row)

	local BarTexture=SM:Fetch("statusbar","BantoBar")

	row.StatusBar:SetStatusBarTexture(BarTexture)
	row.StatusBar:SetStatusBarColor(.5, .5, .5, 1)
	row.StatusBar:SetMinMaxValues(0,100)
	row.StatusBar:SetValue(100)
	row.StatusBar:Show()
	
	row.nameSpace = nameSpace

	row.OnSizeChanged = RowSizeChanged
	row.SetOffset = RowSetOffset
	row.AddColumn = AddColumn
	row.UpdateColumns = UpdateColumns
	row.SetColumn = SetColumn
	row.SetColumnWidth = SetColumnWidth
	row.SetName = SetName
	row.SetNameSpace = SetNameSpace
	row.RemoveColumn = RemoveColumn
	row.RemoveColumns = RemoveColumns
	row.CenterLabels = CenterLabels
	row.MouseDown = Row_MouseDown
	row.MouseUp = Row_MouseUp
	row.MouseOnColumn = Row_MouseOnColumn
	row.SetTooltipFunc = Row_SetTooltipFunc

	row.MouseEnter = Row_MouseEnter
	row.MouseLeave = Row_MouseLeave

	row:SetScript("OnSizeChanged", row.OnSizeChanged)
	row:SetScript("OnMouseDown", row.MouseDown)
	row:SetScript("OnMouseUp", row.MouseUp)
	row:SetScript("OnEnter", row.MouseEnter)
	row:SetScript("OnLeave", row.MouseLeave)
	row:EnableMouse()

	row.Highlight = row.StatusBar:CreateTexture(nil, "OVERLAY")
	row.Highlight:SetTexture(BarTexture)
	row.Highlight:SetAllPoints(row.StatusBar)
	row.Highlight:SetVertexColor(1.0, 1.0, 0.0, 0.33)
	row.Highlight:SetBlendMode("ADD")
	row.Highlight:Hide()

	row.selectedTexture = row.StatusBar:CreateTexture(nil, "OVERLAY")
	row.selectedTexture:SetTexture(BarTexture)
	row.selectedTexture:SetAllPoints(row.StatusBar)
	row.selectedTexture:SetVertexColor(0.0, 1.0, 0.0, 0.33)
	row.selectedTexture:SetBlendMode("ADD")
	row.selectedTexture:Hide()
	
	row.NameText = row.StatusBar:CreateFontString(nil ,"OVERLAY","GameFontHighlightSmall")
	Recount2:AddFontString(row.NameText)
	SetFontSize(row.NameText, rowHeight*0.8)
	row.NameText:SetPoint("LEFT", row, "LEFT", 0, 0)
	row.NameText:SetJustifyH("LEFT")

	if rowWidth<nameSpace then
		row.NameText:SetPoint("RIGHT", row, "LEFT", rowWidth, 0)
	else
		row.NameText:SetPoint("RIGHT", row, "LEFT", nameSpace, 0)
	end

	row.columns=Recount2:GetTable()
	row.columnWidth=Recount2:GetTable()
	row.extraStrings=Recount2:GetTable()

	row.offset = 0

	return row
end

function ShowTooltip(self, rowID)
	Recount2.Widgets:DropdownClear()

	Recount2.Widgets:DropdownAddText("Click for further details")
	Recount2.Widgets:DropdownOpen(true)
end

function Table:UpdateRowCount()
	local numRows = math_floor((self.lastHeight+self.rowSpacing)/(self.rowHeight+self.rowSpacing))

	local self_rows = self.rows

	self.numRows = numRows
	if numRows<#self_rows then
		for rowID = numRows+1, #self_rows do
			self_rows[rowID]:Hide()
		end
	elseif numRows>#self_rows then
		local rowHeight = self.rowHeight
		local rowSpace = self.rowHeight + self.rowSpacing
		local w = self.rowWidth
		for rowID = #self_rows+1, numRows do
			tinsert(self_rows, CreateRow(self, w, rowHeight, self.nameSpace, rowID, self.RowClickedOn))
			self_rows[rowID]:SetPoint("TOPLEFT", self, "TOPLEFT", 0, -(rowID-1)*rowSpace)
			
			if self.RowClickedOn and rowID~=1 then
				self_rows[rowID].highlight = true
				self_rows[rowID]:SetTooltipFunc(ShowTooltip, {self, rowID})
			end
			if self.vScroll:IsShown() then
				self_rows[rowID]:SetPoint("TOPRIGHT", self, "TOPRIGHT", -rowSpace, -(rowID-1)*rowSpace)
			else
				self_rows[rowID]:SetPoint("TOPRIGHT", self, "TOPRIGHT", 0, -(rowID-1)*rowSpace)
			end

			for _, col in ipairs(self.columnsInUse) do
				self_rows[rowID]:AddColumn(self.columns[col].width)
			end
		end
	end
end

function Table:OnSizeChanged()
	local w = self:GetRight()-self:GetLeft()
	local h = self:GetTop()-self:GetBottom()

	if w~=self.lastWidth then
		self.lastWidth = w
		self.colSpace = w - self.nameSpace
		self.hScroll:SetMinMaxValues(0, math_max(self.totalWidth-self.colSpace,0))
		if self.totalWidth<self.colSpace then
			if self.hScroll:IsShown() then
				self.hScroll:Hide()
			end
		elseif not self.hScroll:IsShown() then
			self.hScroll:Show()
		end
		self:DisplayRows()
	end

	if h~=self.lastHeight then
		self.lastHeight = h
		self:UpdateRowCount()
		self:DisplayRows()
	end
end

local sortOn=1

local function ColumnSort(a, b)
	if a[sortOn]==b[sortOn] then
		return a.name<b.name
	end
	return a[sortOn]>b[sortOn]
end

function Table:UpdateRows()
	local data = self.data
	local entryID = 1
	local entry
	local columns = self.columns
	local columnTotal = self.columnTotal
	local sortedData = self.sortedData

	for i, columnNum in ipairs(self.columnsInUse) do
		columnTotal[i] = 0
	end

	for k, v in pairs(data) do
		if sortedData[entryID]==nil then
			sortedData[entryID] = Recount2:GetTable()
		end

		entry = sortedData[entryID]
		entry.name = k
		entry.data = v
		self.entryConstantFunc(k, v)

		for i, columnNum in ipairs(self.columnsInUse) do
			entry[i] = columns[columnNum].value() or 0
			columnTotal[i] = columnTotal[i] + entry[i]
		end

		entryID = entryID + 1
	end

	for i = entryID, #sortedData do
		sortedData[i] = nil
	end

	self.numData = entryID - 1 	

	sortOn = self.sortOn
	table_sort(sortedData, ColumnSort)

	
	self:DisplayRows()
end

function Table:DisplayRows()
	local columns = self.columns
	local columnTotal = self.columnTotal
	local sortedData = self.sortedData
	local numRows = self.numRows

	if self.hScroll:IsShown() then
		numRows = numRows - 1
	end

	if self.numData >=numRows then
		self.vScroll:SetMinMaxValues(0, self.numData-numRows+1)
		if not self.vScroll:IsShown() then
			self.vScroll:Show()
			local rowSpace = self.rowSpacing + self.rowHeight
			self.hScroll:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -rowSpace , 0)
			for i, row in ipairs(self.rows) do
				row:SetPoint("TOPRIGHT", self, "TOPRIGHT", -rowSpace, -(i-1)*rowSpace)
			end
		end
	else
		if self.vScroll:IsShown() then
			self.vScroll:Hide()
			local rowSpace = self.rowSpacing + self.rowHeight
			self.hScroll:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
			for i, row in ipairs(self.rows) do
				row:SetPoint("TOPRIGHT", self, "TOPRIGHT", 0, -(i-1)*rowSpace)
			end
		end
	end

	if self.numData>0 then 
		for i, columnNum in ipairs(self.columnsInUse) do
			Recount2:AddExpressionConstants(columns[columnNum].totalVar, columnTotal[i], columns[columnNum].totalVar.."Avg", columnTotal[i]/self.numData)
		end
	end

	Recount2:AddExpressionConstant('tNum', self.numData)

	local vOffset = math_floor(self.vScroll:GetValue()+0.5)

	if not self.vScroll:IsShown() then
		vOffset = 0
	end

	for k, row in ipairs(self.rows) do
		if k==1 then
			row:SetName("Name")
			for i, columnNum in ipairs(self.columnsInUse) do
				row:SetColumn(i, columns[columnNum].name)
			end
			row.StatusBar:SetValue(0)
			row:CenterLabels()
			row:Show()
		elseif k>numRows then
			row:Hide()
		elseif k<=(self.numData-vOffset+1) then
			local pos = k-1+vOffset
			local entry = sortedData[pos]
			Recount2:AddExpressionConstant('ePos', pos)
			self.entryConstantFunc(entry.name, entry.data)

			row:SetName(Recount2:ProcessCompiledTags(self.rowNameTag))
			row.StatusBar:SetValue(100*entry[self.sortOn]/sortedData[1][self.sortOn])
			for i, columnNum in ipairs(self.columnsInUse) do
				row:SetColumn(i, Recount2:ProcessCompiledTags(columns[columnNum].text))
			end

			if entry.data==self.selected then
				row.selectedTexture:Show()
			else
				row.selectedTexture:Hide()
			end

			row:Show()
		else
			row:Hide()
		end
	end

	self:UpdateColumnPositions()
end

function Table:UpdateColumnPositions()
	local columns = self.columns
	local hOffset = self.hScroll:GetValue()

	local handleOffset = 14

	if not self.hScroll:IsShown() then
		handleOffset = 0
	end

	local handle = self.nameHandle
	if self.nameSpace<self:GetWidth() then
		handle:ClearAllPoints()
		handle:SetPoint("TOPLEFT", self, "TOPLEFT", self.nameSpace-3, 0)
		handle:SetPoint("BOTTOMRIGHT", self, "BOTTOMLEFT", self.nameSpace+3, handleOffset)
		handle:Show()
		handle.texture:Show()
	elseif not handle.moving then
		handle:Hide()
	else
		handle.texture:Hide()
	end

	--Update position of the selected texture
	if #self.columnsInUse>0 then
		local startPos = self.columnPos[self.sortOn] - hOffset
		local endPos = startPos + columns[self.columnsInUse[self.sortOn]].width+4
		local colSpace = self.colSpace

		if self.vScroll:IsShown() then
			colSpace = colSpace - self.rowHeight - self.rowSpacing
		end

		if endPos<0 or startPos>colSpace then
			self.selectedTexture:Hide()
		else
			startPos = self.nameSpace + math_max(startPos, 0)
			endPos = self.nameSpace + math_min(endPos, colSpace)
			self.selectedTexture:ClearAllPoints()
			self.selectedTexture:SetPoint("TOPLEFT", self, "TOPLEFT", startPos, 0)
			self.selectedTexture:SetPoint("BOTTOMRIGHT", self, "BOTTOMLEFT", endPos, 0)
			self.selectedTexture:Show()
		end

		--Update positions of handles
		for i, handle in ipairs(self.columnHandles) do
			endPos = self.columnPos[i] - hOffset + columns[self.columnsInUse[i]].width+4

			if endPos>0 and endPos<colSpace then
				endPos = endPos + self.nameSpace
				handle:ClearAllPoints()
				handle:SetPoint("TOPLEFT", self, "TOPLEFT", endPos-3, 0)
				handle:SetPoint("BOTTOMRIGHT", self, "BOTTOMLEFT", endPos+3, handleOffset)
				handle:Show()
				handle.texture:Show()
			elseif not handle.moving then
				handle:Hide()
			else
				handle.texture:Hide()
			end
		end
	else
		self.selectedTexture:Hide()
	end
	

	for k, row in ipairs(self.rows) do
		row:SetOffset(hOffset)
	end
end

function Table:SetEntryConstantFunc(func)
	self.entryConstantFunc = func
end

function Table:SetData(data)
	self.data = data
end

function Table:SetNameTag(tag)
	self.rowNameTag = Recount2:CompileTags(tag)
end

function Table:AddColumn(columnName, valueExpression, textTag, totalVar, columnWidth)
	local column = Recount2:GetTable()

	column.name = columnName
	column.value = Recount2:CompileExpression(valueExpression)
	column.text = Recount2:CompileTags(textTag)
	column.totalVar = totalVar
	column.width = columnWidth

	tinsert(self.columnPos, self.totalWidth)

	self.totalWidth = self.totalWidth + columnWidth+4
	self.hScroll:SetMinMaxValues(0, math_max(self.totalWidth-self.colSpace+4,0))

	if self.totalWidth<self.colSpace then
		if self.hScroll:IsShown() then
			self.hScroll:Hide()
			self:DisplayRows()
		end
	elseif not self.hScroll:IsShown() then
		self.hScroll:Show()
		self:DisplayRows()
	end

	tinsert(self.columns, column)
	tinsert(self.columnsInUse, #self.columns)

	for _, row in ipairs(self.rows) do
		row:AddColumn(column.width)
	end

	

	handle = CreateFrame("FRAME", nil, self)
	handle:SetWidth(6)
	tinsert(self.columnHandles, handle)
	handle.columnID = #self.columnHandles
	handle.table = self
	handle.OnMouseDown = Table.MouseDownOnHandle
	handle.OnMouseUp = Table.MouseUpOnHandle
	handle.EnterHandle = Table.EnterHandle
	handle.LeaveHandle = Table.LeaveHandle
	handle:SetScript("OnMouseDown", handle.OnMouseDown)
	handle:SetScript("OnMouseUp", handle.OnMouseUp)
	handle:SetScript("OnEnter", handle.EnterHandle)
	handle:SetScript("OnLeave", handle.LeaveHandle)
	handle:EnableMouse()
	handle:SetFrameLevel(self:GetFrameLevel()+2)

	handle.texture = handle:CreateTexture(nil, "BACKGROUND")
	handle.texture:SetTexture(0.35, 0.35, 0.35, 1.0)
	handle.texture:SetWidth(2)
	handle.texture:SetPoint("TOP", handle, "TOP")
	handle.texture:SetPoint("BOTTOM", handle, "BOTTOM")
	handle.texture:Show()
end

function Table:MouseDownOnHandle()
	self.moving = true
	self.table.mouseX, _ = GetCursorPosition()
	Recount2:ScheduleRepeatingTimer("ResizingColumn", Table.UpdateWidth, 0, self.table, self.columnID, self.table.columns[self.table.columnsInUse[self.columnID]])
end

function Table:MouseUpOnHandle()
	self.moving = false
	Recount2:RemoveTimer("ResizingColumn")
end

function Table:MouseDownOnNameHandle()
	self.moving = true
	self.table.mouseX, _ = GetCursorPosition()
	Recount2:ScheduleRepeatingTimer("ResizingName", Table.UpdateNameWidth, 0, self.table)
end

function Table:MouseUpOnNameHandle()
	self.moving = false
	Recount2:RemoveTimer("ResizingName")
end

function Table:EnterHandle()
	SetCursor("Interface\\Addons\\Recount2\\Textures\\Cursors\\cursor_LR_resize.blp")
end

function Table:LeaveHandle()
	ResetCursor()
end


function Table:RemoveColumns()
	for k in ipairs(self.columns) do
		self.columns[k] = nil
	end

	for k in ipairs(self.columnsInUse) do
		self.columnsInUse[k] = nil
	end

	for _, row in ipairs(self.rows) do
		row:RemoveColumns()
	end
end

function Table:SetColumnWidth(id, width)
	for _, row in ipairs(self.rows) do
		row:SetColumnWidth(id, width)
	end

	local curColumn = self.columns[self.columnsInUse[id]]
	local widthChange = width - curColumn.width
	curColumn.width = width
	for i = id+1, #self.columnPos do
		self.columnPos[i] = self.columnPos[i] + widthChange
	end

	self.totalWidth = self.totalWidth + widthChange
	self.hScroll:SetMinMaxValues(0, math_max(self.totalWidth-self.colSpace+4,0))

	if self.totalWidth<self.colSpace then
		if self.hScroll:IsShown() then
			self.hScroll:Hide()
			self:DisplayRows()
		end
	elseif not self.hScroll:IsShown() then
		self.hScroll:Show()
		self:DisplayRows()
	end

	self:UpdateColumnPositions()
end

function Table:UpdateWidth(columnID, column)
	local x, y = GetCursorPosition()

	if x~=self.mouseX then
		self:SetColumnWidth(columnID, math.max(column.width+x-self.mouseX, 25))
		self.mouseX = x
	end
end

function Table:UpdateNameWidth()
	local x, y = GetCursorPosition()

	if x~=self.mouseX then
		local diff = x-self.mouseX
		local prev = self.nameSpace
		self.nameSpace = math.max(self.nameSpace + diff, 50)
		diff = self.nameSpace-prev
		prev = self.colSpace
		self.colSpace = math.max(self.colSpace - diff, 50)
		diff = prev-self.colSpace-diff
		self.nameSpace = self.nameSpace + diff
		for _, row in ipairs(self.rows) do
			row:SetNameSpace(self.nameSpace)
		end
		self.mouseX = x

		--Update H Scroll
		self.hScroll:SetMinMaxValues(0, math_max(self.totalWidth-self.colSpace,0))
		self.hScroll:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", self.nameSpace, 0)
		if self.totalWidth<self.colSpace then
			if self.hScroll:IsShown() then
				self.hScroll:Hide()
			end
		elseif not self.hScroll:IsShown() then
			self.hScroll:Show()
		end

		--Update Colum Positions
		self:UpdateColumnPositions()
	end
end

function Table:RowClickedOn(event, mouseButton, rowID, columnID, columnEdge)
	if event=="OnMouseDown" then
		if mouseButton=="LeftButton" then
			if rowID==1 and columnID and columnID>0 then
				self.sortOn = columnID
				self:UpdateRows()
			elseif rowID>1 and self.RowClickFunc then
				if self.vScroll:IsShown() then
					rowID = (rowID-1)+math_floor(self.vScroll:GetValue()+0.5)
				else
					rowID = (rowID-1)
				end
				self.RowClickFunc(self.rowClickArg, self.sortedData[rowID].data)
			end
		end
	end
end

function Table:MouseWheel(arg1)
	if arg1==1 then
		self.vScroll:Decrease()
	elseif arg1==-1 then
		self.vScroll:Increase()
	end
end

function Table:SetSelection(rowData)
	self.selected = rowData
	self:DisplayRows()
end

function Table:ClearSelection()
	self.selected = nil
end

function Recount2:CreateTableWindow(name, parent, width, height, rowHeight, rowSpacing, nameSpace, rowClickFunc, rowClickArg)
	local tableFrame=CreateFrame("Frame", nil, parent)
	tableFrame:SetWidth(width)
	tableFrame:SetHeight(height)
	tableFrame:SetPoint("CENTER")

	tableFrame.name = name
	tableFrame.lastWidth = width
	tableFrame.lastHeight = height

	tableFrame.rowHeight = rowHeight
	tableFrame.rowSpacing = rowSpacing
	tableFrame.nameSpace = nameSpace
	tableFrame.colSpace = width - nameSpace
	tableFrame.totalWidth = 0
	tableFrame.numData = 0
	tableFrame.sortOn = 1

	local numRows = math_floor((height+rowSpacing)/(rowHeight+rowSpacing))

	tableFrame.numRows = numRows
	tableFrame.RowClickedOn = Table.RowClickedOn
	tableFrame.RowClickFunc = rowClickFunc
	tableFrame.rowClickArg = rowClickArg

	tableFrame.rows = Recount2:GetTable()
	local tableFrame_rows = tableFrame.rows

	local rowSpace = rowSpacing+rowHeight
	tableFrame.rowWidth = width - rowSpace
	local rowWidth = tableFrame.rowWidth
	for i = 1, numRows do
		tinsert(tableFrame_rows, CreateRow(tableFrame, rowWidth, rowHeight, nameSpace, i, tableFrame.RowClickedOn))

		if rowClickFunc and i~=1 then
			tableFrame_rows[i].highlight = true
			tableFrame_rows[i]:SetTooltipFunc(ShowTooltip, {tableFrame, i})
		end
		tableFrame_rows[i]:SetPoint("TOPLEFT", tableFrame, "TOPLEFT", 0, -(i-1)*rowSpace)
		tableFrame_rows[i]:SetPoint("TOPRIGHT", tableFrame, "TOPRIGHT", 0, -(i-1)*rowSpace)
	end

	handle = CreateFrame("FRAME", nil, tableFrame)
	handle:SetWidth(6)
	tableFrame.nameHandle = handle
	handle.OnMouseDown = Table.MouseDownOnNameHandle
	handle.OnMouseUp = Table.MouseUpOnNameHandle
	handle.EnterHandle = Table.EnterHandle
	handle.LeaveHandle = Table.LeaveHandle
	handle:SetScript("OnMouseDown", handle.OnMouseDown)
	handle:SetScript("OnMouseUp", handle.OnMouseUp)
	handle:SetScript("OnEnter", handle.EnterHandle)
	handle:SetScript("OnLeave", handle.LeaveHandle)
	handle:EnableMouse()
	handle:SetFrameLevel(tableFrame:GetFrameLevel()+2)
	handle.table = tableFrame

	handle.texture = handle:CreateTexture(nil, "BACKGROUND")
	handle.texture:SetTexture(0.35, 0.35, 0.35, 1.0)
	handle.texture:SetWidth(2)
	handle.texture:SetPoint("TOP", handle, "TOP")
	handle.texture:SetPoint("BOTTOM", handle, "BOTTOM")
	handle.texture:Show()

	tableFrame.cSort = Recount2:GetTable()

	--tableFrame.OnUpdateRows = Table.OnUpdateRows
	tableFrame.OnSizeChanged = Table.OnSizeChanged
	tableFrame.UpdateRowCount = Table.UpdateRowCount

	tableFrame:SetScript("OnSizeChanged", tableFrame.OnSizeChanged)

	tableFrame.AddColumn = Table.AddColumn
	tableFrame.SetData = Table.SetData
	tableFrame.UpdateRows = Table.UpdateRows
	tableFrame.DisplayRows = Table.DisplayRows
	tableFrame.SetEntryConstantFunc = Table.SetEntryConstantFunc
	tableFrame.RemoveColumns = Table.RemoveColumns
	tableFrame.SetNameTag = Table.SetNameTag
	tableFrame.SetColumnWidth = Table.SetColumnWidth
	tableFrame.UpdateColumnPositions = Table.UpdateColumnPositions

	tableFrame.SetSelection = Table.SetSelection
	tableFrame.ClearSelection = Table.ClearSelection

	tableFrame.columns = Recount2:GetTable()
	tableFrame.columnsInUse = Recount2:GetTable()
	tableFrame.columnPos = Recount2:GetTable()
	tableFrame.columnTotal = Recount2:GetTable()
	tableFrame.columnHandles = Recount2:GetTable()

	tableFrame.sortedData = Recount2:GetTable()


	tableFrame.hScroll = Recount2.Widgets:CreateScrollbar(tableFrame, width-nameSpace-rowSpace, rowHeight)
	tableFrame.hScroll:SetPoint("BOTTOMLEFT", tableFrame, "BOTTOMLEFT", nameSpace, 0)
	tableFrame.hScroll:SetPoint("BOTTOMRIGHT", tableFrame, "BOTTOMRIGHT", -rowSpace, 0)
	tableFrame.hScroll:SetOnChange(tableFrame.UpdateColumnPositions, tableFrame)

	tableFrame.vScroll = Recount2.Widgets:CreateScrollbar(tableFrame, rowHeight, height-rowSpace, "VERTICAL")
	tableFrame.vScroll:SetPoint("TOPRIGHT", tableFrame, "TOPRIGHT")
	tableFrame.vScroll:SetPoint("BOTTOMRIGHT", tableFrame, "BOTTOMRIGHT", 0, rowSpace)
	tableFrame.vScroll:SetOnChange(tableFrame.DisplayRows, tableFrame)
	tableFrame.vScroll:SetValue(0)
	tableFrame.vScroll:SetMinMaxValues(0, 0)


	tableFrame.selectedTexture = tableFrame:CreateTexture(nil, "BACKGROUND")
	tableFrame.selectedTexture:SetTexture(1, 1, 0, 0.25)
	tableFrame.selectedTexture:Hide()

	tableFrame.MouseWheel = Table.MouseWheel
	tableFrame:EnableMouseWheel()
	tableFrame:SetScript("OnMouseWheel", tableFrame.MouseWheel)

	return tableFrame
end